''Module first_fit_c.spin: Heap manager.
'' 
''Copyright (C) 2006 Philip C. Pilgrim
'' 
''This program is free software; you can redistribute it and/or modify
''it under the terms of the GNU General Public License as published by
''the Free Software Foundation; either version 2 of the License, or
''(at your option) any later version.
''
''This program is distributed in the hope that it will be useful,
''but WITHOUT ANY WARRANTY; without even the implied warranty of
''MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
''GNU General Public License for more details.
''
''You should have received a copy of the GNU General Public License
''along with this program; if not, write to the Free Software
''Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
''
''CONTACT
''
''  phil@phipi.com (First contact may require addition to my whitelist.)
''
''VERSION HISTORY
''
''  0.01 alpha: released 2006.04.05
''
''SUMMARY
''
''  This module provides a heap manager for multiple heaps. Heaps can be shared among cogs.
''  It uses the "first fit" technique (algorithm C, "Liberation with Boundary Tags")
''  outlined in Knuth's "Fundamental Algorithms".
''

CON

  MAXHEAPS      = 15                                    'Maximum number of heaps (1 .. 15) that can be managed.

  NXT           = 1
  PRV           = 2
  DATA          = 2
  USED          = $8000

OBJ

  'print :       "print"

PUB start

'' Start the heap manager and assign it a lock. Returns the address of the heap index on success or if already started;
'' returns zero if no lock could be assigned.
  
  'print.start(0, 0, -9600)
  if running                                            'Already running?
    return @heap_base                                   '  Yes: Just say so and leave.
  elseif (master_lock := locknew) => 0                  'Can get a new master lock?
    running := true                                     '  Yes: We're running
    lockclr(master_lock)                                '       Make sure it's unlocked.
    return @heap_base                                   '       And out, reporting success.
  else
    return 0                                            '  No:  Failed to launch.  

PUB new(heap_addr, heap_longs) : heap | link, bytes

'' Create a new heap of heap_longs * 4 bytes in length at byte address heap_addr.
'' Return heap ID on success, zero on failure.

  if not start                                          'Running or can autostart?
    return 0                                            '  No: Too bad.
  heap := 0                                             'Initialize return value. Valid heap address?
  if running and heap_addr and heap_addr == heap_addr & $7FFC and heap_addr + heap_longs << 2 < $8000
    repeat while lockset(master_lock)                   '  Yes: Lock the main gate.
    repeat while heap < MAXHEAPS and heap_users[heap++] > 0 '   Find the next unused heap entry, if any.
    if --heap < MAXHEAPS                                '       Heap index has room?
      longfill(heap_addr, $FFFFFFFF, heap_longs)        '         Yes: Fill the heap with $FF.
      heap_base[heap] := heap_addr                      '              Put the heap address into the table.
      heap_first[heap] := heap_last[heap] := heap_addr += 2 '          Initialize free pointers.
      heap_bytes[heap] := bytes := heap_longs << 2      '              Save the size.
      word[heap_addr] := word[heap_addr + bytes - 6] := bytes - 4 '    Construct the free area.
      word[heap_addr][NXT] := word[heap_addr][PRV] := 0 '              Initialize next and prev pointers.
      heap_users[heap] := 1 << cogid                    '              Add this cog to user list.
      heap_lock &= !(1 << heap)                         '              Make sure heap is unlocked.
      curr_heap[cogid] := ++heap                        '              Return value > 0 on success.
    else                                                
      heap := 0                                         '         No:  Admit defeat.
    lockclr(master_lock)                                '       Unlock the main gate.
  return heap                                           'Leave.

PUB current

'' Get the current heap (top of stack) for this cog.

  return curr_heap[cogid]

PUB push(heap)

'' Push the specified heap (one-based heap index) onto the cog's heap stack, making it the current heap for this cog.
'' Add cog to users' list for the heap if it's not already on there.

  if lock(heap)                                         'Lock heap. Legitimate call?
    curr_heap[cogid] := curr_heap[cogid] << 4 + heap    '  Yes: Push this on the stack as current heap. 
    heap_users[heap - 1] |= 1 << cogid                  '       Add to users' list for this heap.
    unlock(heap)                                        '       Unlock the heap.
    return heap                                         '       Return heap number on success.
  else                                                                             
    return 0                                            '  No:  Return zero on failure.

PUB pop | prev_heap

'' Pop the heap stack. Return the old heap number.

  if running                                            'Legitimate call?
    prev_heap := curr_heap[cogid] & $0F                 '  Yes: Save the current heap number.
    curr_heap[cogid] >>= 4                              '       Pop it off the stack. 
    return prev_heap                                    '       Return the saved number.
  else
    return 0                                            '  No:  Return failure.        
        
PUB abandon(heap) : okay | ptr, top, size

'' Abandon an existing heap. All blocks allocated to this cog will be freed,
'' and all references to this heap in cog's heap stack will be zeroed.
'' If this is the last cog to abandon the heap, the heap will be deleted from the heap index.

  okay := false                                         'Initialize return value.
  if lock(heap--)                                       'Legitimate call?
    if heap_users[heap] &= !(1 << cogid)                '  Yes: Delete cog from user list for this heap. Any left?
      ptr := heap_base[heap] + 2                        '         Yes: Need to free this cog's blocks. Get base address in ptr.
      top := ptr + heap_bytes[heap]                     '              Get limit value.
      repeat while ptr < top                            '              Scan the entire heap:
        size := word[ptr] & $7FFF                       '                Get the size of this block.
        if word[ptr] & USED and word[ptr + size - 2] & $0F == cogid '    Is block allocated to this cog?
          ptr := _dealloc(heap, ptr)                    '                  Yes: Free it, and point to next one.
        else
          ptr += size                                   '                  No:  Next one is the one you'd expect.
      repeat 8                                          '              Eliminate references to this heap from stack:          
        if (curr_heap[cogid] ->= 4) & $0F == (heap + 1) '                Stack entry == heap number?
          curr_heap[cogid] &= $FFFFFFF0                 '                  Yes: Well, we're not using this heap anymore.
    else                                                '         No:  Zero all of heap's entries in index.
      heap_base[heap] := heap_first[heap] := heap_last[heap] := heap_bytes[heap] := 0
    okay := true                                        '       Success!   
    unlock(++heap)                                      '       Unlock the heap.
  return okay                                           'Over and out.

PUB alloc(bytes) : alloc_ptr | heap, cur_free, cur_size

'' Allocate a block of memory on current heap of length "bytes" (rounded up to multiple of four).

  'print.f1(string(" %d:"), bytes)
  alloc_ptr := 0                                        'Initialize return value to "failure".
  if bytes > 0 and bytes < $8000                        'Legitimate request?
    heap := curr_heap[cogid] & $0F                      '  Yes: Get index of current heap.
    if lock(heap--)                                     '       Lock heap. Legitimate call?
      bytes := (((bytes + 4) #> 8) + 3) & $FFFC         '         Yes: Add overhead and alignment to get real size req.
      'print.f1(string("%d:"), bytes)
      cur_free := heap_first[heap]                      '              Initialize "free" link.
      repeat while cur_free                             '              Scan the "free" list:
        cur_size := word[cur_free]                      '                Get size of this block.
        'print.f1(string(" %d"), cur_size)
        if cur_size => bytes                            '                Is block big enough?
          _unfree(cur_free, heap)                       '                  Yes: Take it out of the free chain.
          if cur_size => bytes + 8                      '                       Is the block big enough to split?
            _dofree(cur_free + bytes, cur_size - bytes, heap) '                     Yes: Free the new split-off block.                    
          else
            bytes := cur_size                           '                         No:  Change requirements to fit size.
          word[cur_free] := USED + bytes                '                       Set USED flag and write size.
          word[cur_free + bytes - 2] := $8000 | (heap + 1) << 4 | cogid '       Set USED flag in last word and write owner's ID and heap ID.
          bytefill(cur_free + 2, $00, bytes - 4)        '                       Zero the DATA area, as a courtesy to caller. 
          alloc_ptr := cur_free + DATA                  '                       Point to DATA area of block.
          'print.putc("|")
          quit                                          '                       We're done, so exit the loop.
        cur_free := word[cur_free][NXT]                 '                Repeat with next free block
      unlock(++heap)                                    '              Unlock the heap.
      'print.putc(13)
  return alloc_ptr                                      'We're done here.

PUB dealloc(ptr) : okay | heap, base

'' Deallocate the pointed-to block.

  okay := false                                         'Initialize return value to "failure".
  heap := _heap_id(ptr -= DATA)                                 'Get the heap id embedded in block.
  if lock(heap--)                                       'Lock the heap. Legitimate call?
    if _inheap(heap, ptr)                               '  Yes: Pointer really legit?
      _dealloc(heap, ptr)                               '         Yes: Deallocate the block.     
      okay := true                                      '              We done good!
    unlock(++heap)                                      '       Unlock the heap.
  return okay                                           'Outta here!

PUB put_tag(ptr, tag) : okay | heap, end

'' Tag the pointed-to-block with the 4 lsbs (0 .. 15) of tag.

  okay := false                                         'Initialize return value to zero.
  heap := _heap_id(ptr -= DATA)                                 'Get the heap id embedded in block.
  if lock(heap--)                                       'Lock the heap. Legitimate call?
    if _inheap(heap, ptr)                               '  Yes: Pointer really legit?
      end := ptr + word[ptr] & $7FFF - 2                '         Yes: Get the tag address.
      word[end] := word[end] & $F0FF + (tag & $0F) << 8 '              Insert the tag.     
      okay := true                                      '              We done good!
    unlock(++heap)                                      '       Unlock the heap.
  return okay                                           'Outta here!

PUB get_tag(ptr)

'' Return the tag value (0 .. 15) of the pointed-to block.

  return word[ptr - DATA + word[ptr - DATA] & $7FFF - 2] >> 8 & $0F

PUB heap_id(ptr) : id | heap

'' Return the heap ID of the block pointed to by ptr.

  id := 0
  heap := _heap_id(ptr -= DATA)
  if lock(heap--)
    if _inheap(heap++, ptr)
      id := heap
    unlock(heap)
  return id      

PUB heap_address(heap)

'' Return the base address of the given heap (one-based heap index).

  return heap_base[--heap]

PUB heap_size(heap)

'' Return the size of the heap in bytes (one-based heap index).

  return heap_bytes[--heap]

PUB heap_free(heap, tail)

'' Return the free pointer of the heap (one-based heap index).

  if tail
    return heap_last[--heap]
  else
    return heap_first[--heap]

PUB block_longs(ptr)

'' Return the size of the indicated block in longs.

  return block_bytes(ptr) >> 2

PUB block_words(ptr)

'' Return the size of the indicated block in words.

  return block_bytes(ptr) >> 1

PUB block_bytes(ptr) : size | heap

'' Return the size of the indicated block in bytes.

  size := 0
  heap := _heap_id(ptr -= DATA)                         'Get heap id embedded in block.
  if lock(heap--)                                       'Lock the heap. Legitimate call?
    if _inheap(heap, ptr)                               '  Yes: Pointer legit?
    size := word[ptr] & $7FFF - 4                       '         Yes: Compute the size of the DATA area.
    unlock(++heap)                                      '       Unlock the heap.
  return size                                           'Return size (or zero on failure).

PUB lock(heap) | lock_mask

'' Lock (suspend other operations on) the given heap (one-based heap index).

  if running and heap > 0 and --heap < MAXHEAPS and heap_users[heap]
                                                        'Legitimate call?
    lock_mask := 1 << heap                              '  Yes: Create a mask from the heap number.
    repeat                                              '       Wait for heap to be unlocked...
      repeat while lockset(master_lock)                 '       First, lock the main gate.
      if not heap_lock & lock_mask                      '       Is heap locked?
        quit                                            '         No:  Okay to exit loop and proceed.
      lockclr(master_lock)                              '         Yes: Unlock the main gate and try again.
    heap_lock |= lock_mask                              '       Heap was unlocked, so lock it with main gate still locked.
    lockclr(master_lock)                                '       Now unlock the main gate.   
    return true                                         '       Return success.
  else
    return false                                        '  No:  Return failure.

PUB unlock(heap) | lock_mask

'' Unlock heap (one-based heap index).

  if running and heap > 0 and --heap < MAXHEAPS and heap_users[heap]
                                                        'Legitimate call?
    lock_mask := 1 << heap                              '  Yes: Create a mask from the heap number
    repeat while lockset(master_lock)                   '       Lock the main gate.
    heap_lock &= !lock_mask                             '       Unlock the heap.
    lockclr(master_lock)                                '       Unlock the main gate.   
    return true                                         '       Return success.
  else
    return false                                        '  No:  Return failure.

PRI _dealloc(heap, ptr) | siz, base, p_siz, p_ptr

'' Do the deallocation dirty work on pointer, a la Knuth's algorithm C.
  
  siz := word[ptr] &= $7FFF                             'Get the szie of the block to be freed.
  if not (p_siz := word[ptr - 1]) & USED                'Is the prior adjacent block already free?
    p_ptr := ptr - p_siz                                '  Yes: Remove it from the free list. 
    _unfree(p_ptr, heap)
    siz += word[p_ptr]                                  '       Combine it with this block. 
    ptr := p_ptr
  p_ptr := ptr + siz
  if not (p_siz := word[p_ptr]) & USED                  'Is the next adjacent block already free?
    _unfree(p_ptr, heap)                                '  Yes: Remove it from the free list.
    siz += p_siz                                        '       Combine size with this block.
    p_ptr += p_siz                                      '       Point to next allocated block.                 
  _dofree(ptr, siz, heap)                               'Free the (possibly combined) block.
  return p_ptr                                          'Return the address of the next allocated block.

PRI _dofree(ptr, siz, heap)

'' Free the block.

  word[ptr] := siz                                      'Write the size into first and last words.
  word[ptr + siz - 2] := siz
  word[ptr][NXT] := 0                                   'Appending to end so NXT is zero.
  if word[ptr][PRV] := heap_last[heap]                  'Point to previous block. Was the free list empty?
    word[heap_last[heap]][NXT] := ptr                   '  No:  Point the last free block to this one. 
  else
    heap_first[heap] := ptr                             '  Yes: Point the first block to this one.
  heap_last[heap] := ptr                                'Point "last" to this block.
  'bytefill(ptr + 6, $FF, siz - 8)                       'DEBUG ONLY: Fill with $FF so it shows up on dump.     

PRI _unfree(p_ptr, heap) | p_nxt, p_prv

'' Remove block from free list.

  p_nxt := word[p_ptr][NXT]
  p_prv := word[p_ptr][PRV]
  if p_nxt
    word[p_nxt][PRV] := p_prv
  else
    heap_last[heap] := p_prv
  if p_prv
    word[p_prv][NXT] := p_nxt
  else
    heap_first[heap] := p_nxt

PRI _heap_id(ptr)

'' Return the heap ID of the block pointed to by ptr.

  return word[ptr + word[ptr] & $7FFF - 2] >> 4 & $0F 'Get index of current heap.

PRI _inheap(heap, ptr) | base

'' Return true iff ptr is in the heap (zero-based heap index).

  return (base := heap_base[heap]) > 0 and ptr => base and ptr < base + heap_bytes[heap] and ptr == ptr & $FFFE and word[ptr] & $8000

DAT

'' Individual heap locks.

  heap_lock   long      0

'' Current heap stacks, by cogid.

  curr_heap   long      0[8]

'' Heap index.

  heap_base   word      0[MAXHEAPS]                     'Pointers to beginning of each heap.
  heap_bytes  word      0[MAXHEAPS]                     'Heap sizes.
  heap_first  word      0[MAXHEAPS]                     'Pointers to first free block in each heap.
  heap_last   word      0[MAXHEAPS]                     'Pointers to last free block in each heap.
  heap_users  byte      0[MAXHEAPS]                     'Heap user lists.

'' Heap manager state.
  
  running     byte      false                           'Running flag.
  master_lock byte      -1                              'Master lock number for all heaps.
       