''Module strings: Dynamically-allocated string routines.
'' 
''Copyright (C) 2006 Philip C. Pilgrim
'' 
''This program is free software; you can redistribute it and/or modify
''it under the terms of the GNU General Public License as published by
''the Free Software Foundation; either version 2 of the License, or
''(at your option) any later version.
''
''This program is distributed in the hope that it will be useful,
''but WITHOUT ANY WARRANTY; without even the implied warranty of
''MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
''GNU General Public License for more details.
''
''You should have received a copy of the GNU General Public License
''along with this program; if not, write to the Free Software
''Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
''
''CONTACT
''
''  phil@phipi.com (First contact may require addition to my whitelist.)
''
''VERSION HISTORY
''
''  0.01 alpha: released 2006.04.05
''
''NAME
''
''  strings.spin
''
''SYNOPSIS
''
''  heap_id := strings.start(@Heap_array, HEAP_LONGS)
''
''    where Heap_array is an array of LONGs, and HEAP_LONGS is its size; or
''
''  strings.start_with(heap_id)
''
''    if heap_id is already defined.
''
''  strings.clone(@strg, "X")
''  strings.clone(@strg, string("AB_CD_CD_EFG", 13))
''  strings.substitute(@strg, string("_CD"), "-", 1)
''  print.str(strg)
''
''    prints AB--_EFG
''
''  strings.destroy(@strg)
''
''DESCRIPTION
''
''  This module provides a collection of methods for creating, manipulating,
''  and destroying dynamically-allocated strings and (nested) arrays of strings.
''  It uses the heap object manager "first_fit" to provide the dynamic allocation.
''
''  With any dynamic allocation scheme, the potential for memory leaks always exists.
''  A memory leak occurs when a pointer to an allocated block gets lost, making it
''  impossible to recover the block's space in the heap for reuse. In the worst case
''  enough pointers will be lost that the heap fills up with orphan blocks, rendering
''  further block allocations impossible. The best way to avoid memory leaks is to make
''  sure that all block pointers get assigned to known variable locations. That way,
''  when a variable is assigned a new value, the block previously assigned to it can
''  be reaped for further use.  
''
''  The "strings" object attempts to enforce reusability by ensuring that each new string
''  that it allocates gets assigned to a known location. Methods whose first arguments
''  are "var_addr" or "array_addr" require the address of a variable in which to store
''  a newly-created or modified string or string array. But this requirement, if fully-
''  enforced, has a downside, in that simply returning a string in a function-style call
''  for further use in an expression becomes impossible. And this can lead to some
''  awkward and unwieldy code.
''
''  For this reason, most methods that require a variable address, will let you specify
''  zero for the address. When this happens, the string or array will be assigned to the 
''  object's own internal queue, where it will live a short life -- but long enough to be
''  used in the programmer's current Spin statement. In this case, the method will return
''  the string itself, rather than TRUE or FALSE. But beware: NEVER assign as string created
''  in this fashion to one of your own variables after the fact. It will be deallocated on
''  its own, and you'll be left to wonder what happened to it.
''
''  Other than this caveat, strings created and modified by "strings" can be treated like
''  any other string in Spin. The functions STRSIZE and STRCOMP work as expected, and these
''  strings can be printed using any of the various debug and print objects available.
''
''  Finally, anytime a string argument is required, you can use single-character ASCII
''  literals enclosed in quotes, such as "X", rather than having to say 'string("X")'.
''  The methods figure out what you mean by looking at the value. If it's less than 256,
''  an ASCII literal is assumed. If you need to pass it a string address less than 256,
''  add $10000 to it so it will wrap around the Propeller's memory.     

CON

'' Set the size of the temporary string queue. This represents the maximum number
'' of unassigned strings that can exist in a single Spin statement.

  TEMP_SIZE     = 16
  
VAR

  word  strs[4]                 'Single-character temporary string queue.
  word  temp[TEMP_SIZE]         'Temporary string queue.
  byte  strs_ptr                'Pointer for strs.
  byte  temp_ptr                'Pointer for temp.
  byte  heap_id                 'Heap handle for this instance of strings.

OBJ

  heap :        "first_fit"
 
PUB start(heap_addr, heap_longs)

'' Start with a new heap, given its address and size in LONGs.

  if not heap_id
    heap_id := heap.new(heap_addr, heap_longs)
    heap.pop
  return heap_id

PUB start_with(hp)

'' Start with an existing heap, given the heap ID.

  if not heap_id
    heap_id := hp
  return heap_id

PUB new(var_addr, siz, data)

'' Create a new string of size siz. If data > 255, it's taken as the address of another
'' string, whose contents (up to size bytes) are copied into the new string.
'' If 0 < data < 256, data is taken to be a character, which is used to fill the new string.
'' If dat == 0, an empty string with capacity of size characters is created.
'' On success, if var_addr > 0, the string is assigned to the WORD at that address, any
'' previous string at that address is destroyed, and TRUE is returned;
'' otherwise a pointer to the newly-created string is returned.
'' On failure, FALSE is returned.

  return _assign(var_addr, _new(siz, data))

PUB clone(var_addr, strg)

'' Create a new string which is a copy of strg.

  strg := _str(strg)
  return new(var_addr, strsize(strg), strg)

PUB destroy(var_addr) : okay | strg, j

'' Recursively deallocate an existing string or (nested) array of strings at var_addr.

  okay := false
  if heap.heap_id(strg := word[var_addr]) == heap_id
    if heap.get_tag(strg)
      repeat j from 0 to heap.block_words(strg) - 1
        destroy(strg + j << 1)
    else
      heap.push(heap_id)
      okay :=  heap.dealloc(strg)
      heap.pop
    word[var_addr] := 0
  return okay

PUB append(var_addr, strg)

'' Append a string to the end of a string at the variable given by var_addr.
'' If the variable is already assigned a string, that string is first destroyed.

  if var_addr > 0
    return concat(var_addr, word[var_addr], strg)
  else
    return 0

PUB prepend(var_addr, strg) 

'' Prepend a string to the beginning of a string at the variable given by var_addr.
'' If the variable is already assigned a string, that string is first destroyed.

  if var_addr > 0
    return concat(var_addr, strg, word[var_addr])
  else
    return 0

PUB concat(var_addr, str1, str2) | size1, size2, new_str

'' Create a new string from the concatenation of str2 to str1.
'' If var_addr > 0, assign the new string to the WORD variable at that address,
'' first destroying any string currently assigned to it.

  str1 := _str(str1)
  str2 := _str(str2)
  if new_str := _new((size1 := strsize(str1)) + (size2 := strsize(str2)), str1)
    bytemove(new_str + size1, str2, size2)
  return _assign(var_addr, new_str)

PUB substr(var_addr, strg, begin, len) | new_str

'' Create a new string as a substring of strg, beginning at position start (zero-based)
'' and containing the subsequent len (or fewer if the strg is too short) characters.
'' If begin < 0, start at -begin characters from the end.

  strg := _str(strg)
  if begin < 0
    begin := (begin + strsize(strg)) #> 0 
  return new(var_addr, len <# strsize(strg) - begin #> 0, strg + begin)

PUB remove(var_addr, begin, find, all)

'' Given a string at var_addr (REQUIRED), and beginning at position begin, remove the next
'' occurrence of the string given by find. If all <> 0, remove all occurrences.
'' If begin < 0, start at -begin characters from the end.

  return substitute(var_addr, begin, find, 0, all)

PUB substitute(var_addr, begin, find, repl, all) : n | loc, f_size, r_size

'' Given a string at var_addr (REQUIRED), and beginning at position begin, replace the next
'' occurrence of the string given by find by the string given by replace. If all <> 0, replace all
'' occurrences.
'' If begin < 0, start at -begin characters from the end.

  n := 0
  if var_addr > 0
    find := _str(find)
    f_size := strsize(find)
    repl := _str(repl)
    r_size := strsize(repl)
    repeat
      if (loc := find_next(word[var_addr], begin, find)) => 0
        replace(var_addr, loc, f_size, repl)
        begin := loc + r_size
        n++
    while loc => 0 and all
  return n      
 
PUB insert(var_addr, loc, strg)

'' Insert strg into the string at var_addr (REQUIRED), beginning at loc.
'' If begin < 0, start at -begin characters from the end.

  return replace(var_addr, loc, 0, strg)

PUB delete(var_addr, begin, len)

'' Delete len characters from the string at var_addr (REQUIRED), beginning at position begin.
'' If begin < 0, start at -begin characters from the end.

  return replace(var_addr, begin, len, 0) 

PUB replace(var_addr, begin, len, strg) : new_str | siz, old_size, old_str

'' Replace the substring (indicated by begin and len) in the variable at var_addr (REQUIRED) with strg.
'' If begin < 0, start at -begin characters from the end.

  if var_addr > 0 and (old_str := word[var_addr]) > 0
    strg := _str(strg)
    siz := strsize(strg)
    old_size := strsize(old_str)
    if begin < 0
      begin := (begin + old_size) #> 0
    elseif begin => old_size
      return true
    len <#= old_size - begin
    new_str := _new(old_size - len + siz, 0)
    bytemove(new_str, old_str, begin)
    bytemove(new_str + begin, strg, siz)
    bytemove(new_str + begin + siz, old_str + begin + len, old_size - begin - len)
    return _assign(var_addr, new_str)
  else
    return false

PUB split(array_addr, strg, delim) : n | begin, end, d_size, array

'' Split the string, strg, delimited by the string, delim, into multiple constituent parts.
'' Assign the parts (sans delimiters) to a newly created WORD array, and assign
'' the array address to the WORD at array_addr.
'' Return the number of pieces produced, or zero on failure.

  n := 0
  if array_addr > 0
    heap.push(heap_id)
    array := heap.alloc((count(strg, delim) + 1) << 1)
    heap.put_tag(array, 1)
    heap.pop
    strg := _str(strg)
    delim := _str(delim)
    d_size := strsize(delim)
    begin := 0
    repeat
      if (end := find_next(strg, begin, delim)) => 0
        substr(array + n++ << 1, strg, begin, end - begin)
        begin := end + d_size
      else
        substr(array + n++ << 1, strg, begin, strsize(strg) - begin)
        quit
    _assign(array_addr, array)
  return n

PUB join(var_addr, array_addr, delim) | j, n, siz, d_size, new_str, ptr, a_size

'' Given an array of n pointers to strings, join them into one string using delim as a
'' delimiter. If var_addr > 0, assign the string to the WORD there and return TRUE; else
'' return a pointer to the string.

  delim := _str(delim)
  if type(array_addr) == 2 and n := size(array_addr)
    siz := (n - 1) * (d_size := strsize(delim))
    repeat j from 0 to (n - 1) << 1 step 2
      siz += strsize(word[array_addr + j])
    ptr := new_str := _new(siz, 0)
    repeat j from 0 to (n - 1) << 1 step 2
      bytemove(ptr, word[array_addr + j], a_size := strsize(word[array_addr + j]))
      if (ptr += a_size) < new_str + siz
        bytemove(ptr, delim, d_size)
        ptr += d_size
  else
    new_str := _str(0)
  return _assign(var_addr, new_str)
    

PUB count(strg, targ_str) : n | begin, end, found, d_size

'' Count the number of occurrences of targ_str in strg.

  n := 0
  strg := _str(strg)
  targ_str := _str(targ_str)
  d_size := strsize(targ_str)
  begin := 0
  repeat
    if (end := find_next(strg, begin, targ_str)) => 0
      begin := end + d_size
      n++
    else
      quit
  return n 

PUB find_next(strg, begin, targ_str) : found | siz, n, targ_size, pos, ch

'' Look for targ_str in string or array given by strg, beginning at begin and searching forward.
'' If begin < 0, start at -begin elements from the end.
'' Return the position of the first character of the substring in strg matching targ_str.
'' Returns -1 if no such substring was found.

  found := -1
  targ_str := _str(targ_str)
  if heap.get_tag(strg)
    siz := size(strg)
    if begin < 0
      begin := begin + siz #> 0
    if begin < siz  
      repeat n from begin to siz - 1
        if strcomp(word[strg][n], targ_str)
          found := n
          quit    
  else    
    siz := strsize(strg)
    if begin < 0
      begin := begin + siz #> 0
    if (n := siz - begin - (targ_size := strsize(targ_str))) => 0
      repeat pos from strg + begin to strg + begin + n
        if _match(pos, targ_str, targ_size)
          found := pos - strg
          quit
  return found

PUB find_prev(strg, begin, targ_str) : found | siz, n, targ_size, pos, ch
    
'' Look for targ_str in strg, beginning at begin and searching backward.
'' If begin < 0, start at -begin characters from the end.
'' Return the position of the first character of the substring in strg matching targ_str.
'' Returns -1 if no such substring was found.

  found := -1
  targ_str := _str(targ_str)
  if begin < 0
    begin := begin + siz #> 0
  if heap.get_tag(strg)
    begin <#= heap.block_words(strg) - 1  
    repeat n from begin to 0
      if strcomp(word[strg][n], targ_str)
        found := n
        quit    
  else
    siz := strsize(strg)
    if (begin <#= siz - (targ_size := strsize(targ_str))) => 0
      pos := strg + begin + 1
      repeat while --pos => strg
        if _match(pos, targ_str, targ_size)
          found := pos - strg
          quit
  return found

PUB get_dict(dict, look_for) : def | entry

'' The "dictionary", dict, is an array of "entries", or pairs of items. The first item in each entry is
'' taken to be a string, the "word" being defined; the second item, the "definition", can be anything.
'' The argument look_for is the string being looked up in the dictionary. The dictionary entries are searched
'' for a word that matches look_for. If one is found, its definition is returned; otherwise, zero is returned.

  if entry := find_dict(dict, look_for)
    return word[entry][1]
  else
    return 0

PUB put_dict(dict, look_for, def) | entry

'' The definition in dict corresponding to look_for is replaced with def. If wrd is not found, no entry
'' is created for it, and FALSE is returned.

  if entry := find_dict(dict, look_for)
    return clone(@word[entry][1], def)
  else
    return false    
  
PUB find_dict(dict, look_for) : entry | index

'' The dictionary, dict, is searched for an entry matching look_for. If one is found, its index in the dictionary
'' is returned; otherwise -1 is returned.

  entry := -1
  if type(dict) == 2
    repeat index from 0 to size(dict) - 1
      entry := word[dict][index]
      if strcomp(word[entry][0], look_for)
        return entry
    entry := 0
  return entry      

PUB size(strg) : siz | j

'' If strg is a string, return STRSIZE for the string. If strg is a dynamically-allocated WORD array, return
'' the number of entries. Otherwise return zero.

  if strg > 0
    if heap.get_tag(strg)
      siz := heap.block_words(strg)
      repeat until word[strg][siz - 1] or siz == 0
        siz--
      return siz
    else
      return strsize(strg)
  else
    return 0

PUB type(strg)

'' If strg is zero, return 0; if a dynamically-allocated string, return 1; if a dynamically-allocated array, return 2.
'' Otherwise, the return value is unspecified.

  if strg > 0
    return heap.get_tag(strg) + 1
  else
    return 0

PUB uc(strg, all) | j, lim, ch

'' if all == 0, change the first character in strg to upper case; if all <>0, change all characters to upper case.

  strg := _str(strg)
  if all
    lim := strsize(strg) - 1
  else
    lim := 0
  repeat j from 0 to lim
    ch := byte[strg][j]
    if ch => "a" and ch =< "z"
      ch -= $20
      byte[strg][j] := ch
  return strg      

PUB lc(strg, all) | j, lim, ch

'' if all == 0, change the first character in strg to lower case; if all <>0, change all characters to lower case.

  strg := _str(strg)
  if all
    lim := strsize(strg) - 1
  else
    lim := 0
  repeat j from 0 to lim
    ch := byte[strg][j]
    if ch => "A" and ch =< "Z"
      ch += $20
      byte[strg][j] := ch
  return strg      

PRI _new(siz, data) : new_str

'' Create a new string of size siz. If data > 255, assume it's a string address, and copy at most siz bytes of data
'' to the new string. If data =< 255, fill the new string with the byte given by data. Return a pointer to the new string.

  new_str := 0
  if siz > 0
    heap.push(heap_id)
    if new_str := heap.alloc(siz + 1)
      if data > 255
        bytemove(new_str, data, siz <# strsize(data))
      elseif data
        bytefill(new_str, data, siz)
    heap.pop  
  return new_str      

PRI _str(strg) | new_str

'' If strg is a single character, create a temporary string from it and return a pointer to
'' the temporary string. Otherwise just return the original argument.

  if strg < 256
    strs[strs_ptr] := strg
    new_str := @strs[strs_ptr++]
    strs_ptr &= 3 
    return new_str
  else
    return strg

PRI _assign(var_addr, strg) : ret_val

'' Assign string strg to var_addr if var_addr > 0, else to a temporary variable in the temp queue.

  ret_val := true
  if var_addr =< 0
    var_addr := @temp[temp_ptr++]
    temp_ptr &= constant(|< >| (TEMP_SIZE >> 1) - 1)
    ret_val := strg
  destroy(var_addr)
  word[var_addr] := strg
  return ret_val              

PRI _match(strg, targ_str, targ_size) : match | end, ch

'' Return true iff the substring of size targ_size, beginning at strg matches targ_str.

  end := strg + targ_size
  ch := byte[end]
  byte[end] := 0
  match := strcomp(strg, targ_str)
  byte[end] := ch
  return match
  